from abc import ABC, abstractmethod

class AbstractLlmService(ABC):
    """
    Abstract class for different Language Learning Model (LLM) Providers.
    """

    @abstractmethod
    def embeddings(self, text: str) -> list:
        """
        Fetches embeddings for the given text. This method should be implemented by subclasses 
        to use the specific LLM provider's embeddings API.
        """
        raise NotImplementedError

    @abstractmethod
    def chat_completion(self, messages, model, temperature = None) -> str:
        """
        Fetches a chat completion response. This method should be implemented by subclasses 
        to interact with the specific LLM provider's chat completion API.
        """
        raise NotImplementedError

    @abstractmethod
    def chat_completion_json(self, messages, model, temperature = None) -> str:
        """
        Fetches a chat completion response in JSON format. This method should be implemented 
        by subclasses to interact with the specific LLM provider's chat completion API using JSON mode.
        """
        raise NotImplementedError

    @abstractmethod
    def json_completion(self, messages, model, temperature = None):
        """
        Helper method to be implemented by subclasses. This method should call either chat_completion or chat_completion_json.
        
        """
        raise NotImplementedError



import re
import json
import requests
from typing import Any, Optional

def _standardize_url(url):
    """
    Generate a standardized URL by adding 'http://' if it's missing.

    Args:
        url (str): The input URL to be standardized.

    Returns:
        str: The standardized URL.
    """
    if url.startswith("http://") or url.startswith("https://"):
        return url
    else:
        return "http://" + url

def _preprocess_strings(keywords, response, case_sensitive):
    """
    Preprocess the keywords based on the case_sensitive flag.

    Args:
        keywords (str or List[str]): The keyword(s) to preprocess.
        case_sensitive (bool): Whether the preprocessing should be case-sensitive.

    Returns:
        List[str]: The preprocessed keywords.
    """
    # If keywords is a string, convert it to a list
    if isinstance(keywords, str):
        keywords = keywords.split(",")

    # Strip leading and spaces from the keywords
    keywords = list(map(lambda k: k.strip(), keywords))

    # If case_sensitive is False, convert all keywords and response to lowercase
    if not case_sensitive:
        keywords = [keyword.lower() for keyword in keywords]
        response = response.lower()

    return keywords, response

def regex(pattern, response):
    """
    Perform a regex search on the response and return a dictionary indicating whether the pattern was found.

    Args:
        pattern (str): The regex pattern to search for.
        response (str): The response string to search within.

    Returns:
        dict: A dictionary containing the result of the regex search and the reason for the result.
    """
    match = re.search(pattern, response)
    if match:
        return {"result": True, "reason": f"regex pattern {pattern} found in output"}
    else:
        return {
            "result": False,
            "reason": f"regex pattern {pattern} not found in output",
        }

def contains_any(keywords, response, case_sensitive=False):
    """
    Check if any of the provided keywords are present in the response.

    Args:
        keywords (str or List[str]): The keyword(s) to search for in the response.
        response (str): The response string to search within.
        case_sensitive (bool, optional): Whether the search should be case-sensitive. Defaults to False.

    Returns:
        dict: A dictionary containing the result of the search and the reason for the result.
    """
    keywords, response = _preprocess_strings(keywords, response, case_sensitive)
    found_keywords = []
    for keyword in keywords:
        if keyword in response:
            found_keywords.append(keyword)

    if found_keywords:
        result = True
        reason = f"One or more keywords were found in output: " + ", ".join(
            found_keywords
        )
    else:
        result = False
        reason = "No keywords found in output"

    return {"result": result, "reason": reason}

def contains_all(keywords, response, case_sensitive=False):
    """
    Check if all the provided keywords are present in the response.

    Args:
        keywords (List[str]): The list of keywords to search for in the response.
        response (str): The response string to search within.
        case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.

    Returns:
        dict: A dictionary containing the result of the keyword search and the reason for the result.
    """
    keywords, response = _preprocess_strings(keywords, response, case_sensitive)
    missing_keywords = []
    for keyword in keywords:
        if keyword not in response:
            result = False
            missing_keywords.append(keyword)
    if (len(missing_keywords)) > 0:
        result = False
        reason = f"keywords not found in output: " + ", ".join(missing_keywords)
    else:
        result = True
        reason = f"{len(keywords)}/{len(keywords)} keywords found in output"

    return {"result": result, "reason": reason}

def contains(keyword, response, case_sensitive=False):
    """
    Check if the response contains a specific keyword.

    Args:
        keyword (str): The keyword to search for in the response.
        response (str): The response string to search within.
        case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.

    Returns:
        dict: A dictionary containing the result of the keyword search and the reason for the result.
    """
    if case_sensitive == False:
        response = response.lower()
        keyword = keyword.lower()
    if keyword not in response:
        result = False
        reason = f"keyword not found in output: " + keyword
    else:
        result = True
        reason = f"keyword {keyword} found in output"

    return {"result": result, "reason": reason}

def contains_none(keywords, response, case_sensitive=False):
    """
    Check if none of the provided keywords are present in the response.

    Args:
        keywords (str or List[str]): The keyword(s) to search for in the response.
        response (str): The response string to search within.
        case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.

    Returns:
        dict: A dictionary containing the result of the check and the reason for the result.
    """
    keywords, response = _preprocess_strings(keywords, response, case_sensitive)
    found_keywords = []
    for keyword in keywords:
        if keyword in response:
            found_keywords.append(keyword)

    if found_keywords:
        result = False
        reason = f"One or more keywords were found in output: " + ", ".join(
            found_keywords
        )
    else:
        result = True
        reason = "No keywords found in output"

    return {"result": result, "reason": reason}

def contains_json(response):
    """
    Check if the response contains valid JSON.

    Args:
        response (str): The response string to check for valid JSON.

    Returns:
        dict: A dictionary containing the result of the JSON check and the reason for the result.
    """
    trimmed_output = response.strip()
    pattern = r'\{(?:\s*"(?:\\.|[^"\\])*"\s*:\s*(?:"(?:\\.|[^"\\])*"|[^{}\[\]:,]+)|[^{}]+)*\}'
    matches = re.findall(pattern, trimmed_output)

    if matches:
        results = []
        errors = []
        for potential_json_string in matches:
            try:
                parsed_json = json.loads(potential_json_string)
                results.append({"json": parsed_json, "valid": True})
            except json.JSONDecodeError as e:
                errors.append({"json": potential_json_string, "valid": False, "error": str(e)})
        if errors:
            return {"result": False, "reason": "Output contains a potential JSON but it is invalid", "matches": results, "errors": errors}
        else:   
            return {"result": True, "reason": "Output contains JSON", "matches": results}
    else:
        return {"result": False, "reason": "Output does not contain JSON"}
    
def contains_email(response):
    """
    Check if the response contains an email address.

    Args:
        response (str): The response string to check for an email address.

    Returns:
        dict: A dictionary containing the result of the email address check and the reason for the result.
    """
    return regex(r'[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+', response)

def is_json(response):
    """
    Check if the response contains valid JSON.

    Args:
        response (str): The response string to check for valid JSON.

    Returns:
        dict: A dictionary containing the result of the JSON check and the reason for the result.
    """
    try:
        json.loads(response)
        result = True
    except json.JSONDecodeError:
        result = False
    if result:
        return {
            "result": True,
            "reason": "Output contains JSON",
        }
    else:
        return {
            "result": False,
            "reason": "Output does not contain JSON",
        }

def is_email(response):
    """
    Check if the response is a valid email address.

    Args:
        response (str): The response string to check for a valid email address.

    Returns:
        dict: A dictionary containing the result of the email address check and the reason for the result.
    """
    return regex(r'^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$', response)

def contains_link(response):
    """
    Check if the response contains a link.

    Args:
        response (str): The response string to check for a link.

    Returns:
        dict: A dictionary containing the result of the link check and the reason for the result.
    """
    pattern = r"(?!.*@)(?:https?://)?(?:www\.)?\S+\.\S+"
    result = bool(re.search(pattern, response))
    if result:
        return {"result": True, "reason": "Link found in output"}
    else:
        return {"result": False, "reason": "No link found in output"}

def contains_valid_link(response):
    """
    Check if the response contains a valid link.

    Args:
        response (str): The response string to check for a valid link.

    Returns:
        dict: A dictionary containing the result of the link check and the reason for the result.
    """
    pattern = r"(?!.*@)(?:https?://)?(?:www\.)?\S+\.\S+"
    link_match = re.search(pattern=pattern, string=response)
    if link_match:
        matched_url = link_match.group()
        if matched_url:
            standardized_url = _standardize_url(matched_url)
            try:
                response = requests.head(standardized_url)
                if response.status_code == 200:
                    return {
                        "result": True,
                        "reason": f"link {matched_url} found in output and is valid",
                    }
                else:
                    return {
                        "result": False,
                        "reason": f"link {matched_url} found in output but is invalid",
                    }
            except:
                return {
                    "result": False,
                    "reason": f"link {matched_url} found in output but is invalid",
                }
    return {"result": False, "reason": f"no link found in output"}

def no_invalid_links(response):
    """
    Check for invalid links in the response.

    Args:
        response (str): The response string to check for invalid links.

    Returns:
        dict: A dictionary containing the result of the link check and the reason for the result.
    """
    pattern = r"(?!.*@)(?:https?://)?(?:www\.)?\S+\.\S+"
    link_match = re.search(pattern=pattern, string=response)
    if link_match:
        matched_url = link_match.group()
        if matched_url:
            standardized_url = _standardize_url(matched_url)
            try:
                response = requests.head(standardized_url)
                if response.status_code == 200:
                    return {
                        "result": True,
                        "reason": f"link {matched_url} found in output and is valid",
                    }
                else:
                    return {
                        "result": False,
                        "reason": f"link {matched_url} found in output but is invalid",
                    }
            except:
                return {
                    "result": False,
                    "reason": f"link {matched_url} found in output but is invalid",
                }
    return {"result": True, "reason": f"no invalid link found in output"}

def api_call(
    url: str,
    response: str,
    query: Optional[str] = None,
    context: Optional[str] = None,
    expected_response: Optional[str] = None,
    payload: dict = None,
    headers: dict = None,
):
    """
    Make an API call with payload to the specified URL.

    Args:
        url (str): The URL to make the API call to.
        response (str): The response to be added to the payload.
        query (Optional[str]): The query parameter to be added to the payload.
        context (Optional[str]): The context parameter to be added to the payload.
        expected_response (Optional[str]): The expected response parameter to be added to the payload.
        payload (dict, optional): The payload to be sent in the API call. Defaults to None.
        headers (dict, optional): The headers to be included in the API call. Defaults to None.

    Returns:
        dict: A dictionary containing the result and reason of the API call.
    """
    if payload is None:
        payload = {}
    if headers is None:
        headers = {}
    payload["response"] = response
    if query:
        payload["query"] = query
    if context:
        payload["context"] = context
    if expected_response:
        payload["expected_response"] = expected_response
    # Check the status code and set the reason accordingly
    try:
        api_response = requests.post(url, json=payload, headers=headers)
        if api_response.status_code == 200:
            # Success
            result = api_response.json().get("result")
            reason = api_response.json().get("reason")
        elif api_response.status_code == 400:
            # Bad Request
            result = False
            reason = "Bad Request: The server could not understand the request due to invalid syntax."
        elif api_response.status_code == 401:
            # Unauthorized
            result = False
            reason = "Unauthorized: Authentication is required and has failed or has not been provided."
        elif api_response.status_code == 500:
            # Internal Server Error
            result = False
            reason = "Internal Server Error: The server encountered an unexpected condition."
        else:
            # Other error codes
            result = False
            reason = f"An error occurred: {api_response.status_code}"
    except Exception as e:
        # Handle any exceptions that occur during the API call
        result = False
        reason = f"API Request Exception: {e}"
        
    return {
        "result": result,
        "reason": reason
    }

def equals(expected_response, response, case_sensitive=False):
    """
    Check if the response exactly matches the expected response.

    Args:
        expected_response (str): The expected response to compare against.
        response (str): The response to compare with the expected output.
        case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.

    Returns:
        dict: A dictionary containing the result and reason of the comparison.
    """
    if case_sensitive == False:
        response = response.lower()
        expected_response = expected_response.lower()
    if response == expected_response:
        result = True
        reason = "✅ output exactly matches expected response"
    else:
        result = False
        reason = "output does not exactly match expected response"
    return {"result": result, "reason": reason}

def starts_with(substring, response, case_sensitive=False):
    """
    Check if the response starts with a specified substring.

    Args:
        substring (str): The substring to check for at the start of the response.
        response (str): The response string to check.
        case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.

    Returns:
        dict: A dictionary containing the result of the check and the reason for the result.
    """
    if case_sensitive == False:
        response = response.lower()
        substring = substring.lower()
    result = response.startswith(substring)
    if result == True:
        return {"result": result, "reason": "output starts with " + substring}
    else:
        return {"result": result, "reason": "output does not start with " + substring}

def ends_with(substring, response, case_sensitive=False):
    """
    Check if the response ends with a specified substring.

    Args:
        substring (str): The substring to check for at the end of the response.
        response (str): The response string to check.
        case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.

    Returns:
        dict: A dictionary containing the result of the check and the reason for the result.
    """
    if case_sensitive == False:
        response = response.lower()
        substring = substring.lower()
    result = response.endswith(substring)
    if result == True:
        return {"result": result, "reason": "output ends with " + substring}
    else:
        return {"result": result, "reason": "output does not end with " + substring}

def length_less_than(max_length, response):
    """
    Check if the length of the response is less than a specified maximum length.

    Args:
        max_length (int): The maximum length that the response should have.
        response (str): The response string to check the length of.

    Returns:
        dict: A dictionary containing the result of the length check and the reason for the result.
    """
    if len(response) < max_length:
        return {
            "result": True,
            "reason": f"output length is less than {max_length} characters",
        }
    else:
        return {
            "result": False,
            "reason": f"output length is greater than {max_length} characters",
        }

def length_greater_than(min_length, response):
    """
    Check if the length of the response is greater than a specified minimum length.

    Args:
        min_length (int): The minimum length that the response should have.
        response (str): The response string to check the length of.

    Returns:
        dict: A dictionary containing the result of the length check and the reason for the result.
    """
    if len(response) > min_length:
        return {
            "result": True,
            "reason": f"output length is greater than {min_length} characters",
        }
    else:
        return {
            "result": False,
            "reason": f"output length is less than {min_length} characters",
        }

"""
A dictionary containing the available operations and their corresponding functions.
"""
operations = {
    "Regex": regex,
    "ContainsAny": contains_any,
    "ContainsAll": contains_all,
    "Contains": contains,
    "ContainsNone": contains_none,
    "ContainsJson": contains_json,
    "ContainsEmail": contains_email,
    "IsJson": is_json,
    "IsEmail": is_email,
    "NoInvalidLinks": no_invalid_links,
    "ContainsLink": contains_link,
    "ContainsValidLink": contains_valid_link,
    "Equals": equals,
    "StartsWith": starts_with,
    "EndsWith": ends_with,
    "LengthLessThan": length_less_than,
    "LengthGreaterThan": length_greater_than,
    "ApiCall": api_call,
}


from typing import List, TypedDict, Optional
from athina.helpers.athina_logging_helper import AthinaLoggingHelper
from athina.evals.llm.llm_evaluator import LlmEvaluator
from athina.evals.base_evaluator import BaseEvaluator
from athina.interfaces.result import EvalResult, BatchRunResult
from athina.interfaces.data import DataPoint
from athina.interfaces.athina import AthinaExperiment
from athina.services.athina_api_service import AthinaApiService
import pandas as pd
import json
import hashlib


class DataPointWithEvalResults(TypedDict):
    """A data point with its evaluation results."""

    data_point: DataPoint
    eval_results: List[EvalResult]


class LlmEvaluatorDescription(TypedDict):
    """A description of an LLM evaluator."""

    name: str
    display_name: str


class LlmBatchEvalResult(TypedDict):
    """Result of running a batch of LLM evaluations."""

    results: List[EvalResult]
    total_runtime: float
    passed_evals: int
    failed_evals: int
    total_evals: int
    total_datapoints: int


class EvalRunner:
    @staticmethod
    def eval_results_link(eval_request_id: str):
        return f"https://app.athina.ai/develop/request/{eval_request_id}"

    @staticmethod
    def flatten_eval_results(batch_eval_results) -> List:
        # Flatten the list of lists into a single list of evaluation results
        flattened_results = [
            item for sublist in batch_eval_results for item in (sublist if sublist is not None else [None])
        ]
        return flattened_results
    
    @staticmethod
    def _create_eval_request(eval_suite_name: str, data) -> Optional[str]:
        try:
            eval_request_id = AthinaLoggingHelper.create_eval_request(
                eval_name=eval_suite_name,
                request_data={"data": data},
                request_type="suite",
            )
            return eval_request_id
        except Exception as e:
            return None
    
    @staticmethod
    def _log_experiment(experiment, eval_request_id: Optional[str]):
        try:
            if experiment is not None and eval_request_id is not None:
                AthinaLoggingHelper.log_experiment(
                    eval_request_id=eval_request_id,
                    experiment=experiment,
                )
        except Exception as e:
            pass

    @staticmethod
    def _log_evaluation_results(eval_results: List[Optional[EvalResult]], eval_request_id: Optional[str]):
        if eval_request_id:
                try:
                    AthinaLoggingHelper.log_eval_results(
                        eval_request_id=eval_request_id,
                        eval_results=eval_results,
                    )
                except Exception as e:
                    pass

    @staticmethod
    def to_df(batch_eval_results):
        # Initialize a dictionary to hold the aggregated data
        aggregated_data = {}

        flattened_results = EvalRunner.flatten_eval_results(
            batch_eval_results=batch_eval_results
        )
        # Process each evaluation result
        for eval_result in flattened_results:
            if eval_result is not None:
                # Serialize and hash the datapoint dictionary to create a unique identifier
                datapoint_hash = hashlib.md5(
                    json.dumps(eval_result["data"], sort_keys=True).encode()
                ).hexdigest()

                # Initialize the datapoint in the aggregated data if not already present
                if datapoint_hash not in aggregated_data:
                    aggregated_data[datapoint_hash] = eval_result["data"]  # Include datapoint details

                # Update the aggregated data with metrics from this evaluation
                for metric in eval_result["metrics"]:
                    metric_name = metric["id"]
                    metric_value = metric["value"]
                    aggregated_data[datapoint_hash][eval_result["display_name"] + " " + metric_name] = metric_value

        # Convert the aggregated data into a DataFrame
        df = pd.DataFrame(list(aggregated_data.values()))

        return df


    @staticmethod
    def run_suite(
        evals: List[BaseEvaluator],
        data: List[DataPoint],
        experiment: Optional[AthinaExperiment] = None,
        max_parallel_evals: int = 1,
    ) -> List[LlmBatchEvalResult]:
        """
        Run a suite of LLM evaluations against a dataset.

        Args:
            evals: A list of LlmEvaluator objects.
            data: A list of data points.

        Returns:
            A list of LlmBatchEvalResult objects.
        """
        eval_suite_name = "llm_eval_suite" + "_" + ",".join(eval.name for eval in evals)
        AthinaApiService.log_usage(eval_name=eval_suite_name, run_type="suite")
        # Create eval request
        eval_request_id = EvalRunner._create_eval_request(eval_suite_name=eval_suite_name, data=data)
        EvalRunner._log_experiment(eval_request_id=eval_request_id, experiment=experiment)

        batch_results = []
        for eval in evals:
            # Run the evaluations
            if max_parallel_evals > 1:
                eval_results = eval._run_batch_generator_async(data, max_parallel_evals)
            else:
                eval_results = list(eval._run_batch_generator(data))

            EvalRunner._log_evaluation_results(eval_request_id=eval_request_id, eval_results=eval_results)
            batch_results.append(eval_results)

        print(
            f"You can view the evaluation results at {EvalRunner.eval_results_link(eval_request_id)}"
        )
        return EvalRunner.to_df(batch_results)




from typing import Optional, List
from athina.evals.eval_type import FunctionEvalTypeId
from athina.evals.function.function_evaluator import FunctionEvaluator


class ContainsAny(FunctionEvaluator):
    def __init__(
        self,
        keywords: List[str],
        case_sensitive: Optional[bool] = False,
    ):
        """
        Initialize the ContainsAny function evaluator.

        Args:
            keywords (List[str]): List of keywords to check for in the response.
            case_sensitive (Optional[bool], optional): Whether the keyword matching should be case sensitive. Defaults to False.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.CONTAINS_ANY.value,
            function_arguments={"keywords":keywords, "case_sensitive":case_sensitive},
        )

class Regex(FunctionEvaluator):
    def __init__(
        self,
        regex: str,
    ):
        """
        Initialize the Regex function evaluator.

        Args:
            regex (str): The regular expression pattern to be matched in the response.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.REGEX.value,
            function_arguments={"pattern":regex},
        )

class ContainsNone(FunctionEvaluator):
    def __init__(self, keywords: List[str], case_sensitive: bool = False):
        """
        Initialize the ContainsNone function evaluator.

        Args:
            keywords (str or List[str]): The keyword(s) to search for in the response.
            case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.CONTAINS_NONE.value,
            function_arguments={
                "keywords": keywords,
                "case_sensitive": case_sensitive,
            },
        )

class Contains(FunctionEvaluator):
    def __init__(self, keyword: str, case_sensitive: bool = False):
        """
        Initialize the Contains function evaluator.

        Args:
            keyword (str): The keyword to search for in the response.
            case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.CONTAINS.value,
            function_arguments={
                "keyword": keyword,
                "case_sensitive": case_sensitive,
            },
        )

class ContainsAll(FunctionEvaluator):
    def __init__(self, keywords: List[str], case_sensitive: bool = False):
        """
        Initialize the ContainsAll function evaluator.

        Args:
            keywords (List[str]): The list of keywords to search for in the response.
            case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.CONTAINS_ALL.value,
            function_arguments={
                "keywords": keywords,
                "case_sensitive": case_sensitive,
            },
        )

class ContainsJson(FunctionEvaluator):
    def __init__(self):
        """
        Initialize the ContainsJson function evaluator.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.CONTAINS_JSON.value,
            function_arguments={},
        )

class ContainsEmail(FunctionEvaluator):
    def __init__(self):
        """
        Initialize the ContainsEmail function evaluator.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.CONTAINS_EMAIL.value,
            function_arguments={},
        )

class IsJson(FunctionEvaluator):
    def __init__(self):
        """
        Initialize the IsJson function evaluator.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.IS_JSON.value,
            function_arguments={},
        )

class IsEmail(FunctionEvaluator):
    def __init__(self):
        """
        Initialize the IsEmail function evaluator.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.IS_EMAIL.value,
            function_arguments={},
        )

class NoInvalidLinks(FunctionEvaluator):
    def __init__(self):
        """
        Initialize the NoInvalidLinks function evaluator.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.NO_INVALID_LINKS.value,
            function_arguments={},
        )

class ContainsLink(FunctionEvaluator):
    def __init__(self):
        """
        Initialize the ContainsLink function evaluator.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.CONTAINS_LINK.value,
            function_arguments={},
        )

class ContainsValidLink(FunctionEvaluator):
    def __init__(self):
        """
        Initialize the ContainsValidLink function evaluator.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.CONTAINS_VALID_LINK.value,
            function_arguments={},
        )

class Equals(FunctionEvaluator):
    def __init__(self, expected_response: str, case_sensitive: bool = False):
        """
        Initialize the Equals function evaluator.

        Args:
            expected_response (str): The expected response to compare against.
            case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.EQUALS.value,
            function_arguments={
                "expected_response": expected_response,
                "case_sensitive": case_sensitive,
            },
        )

class StartsWith(FunctionEvaluator):
    def __init__(self, substring: str, case_sensitive: bool = False):
        """
        Initialize the StartsWith function evaluator.

        Args:
            substring (str): The substring to check for at the start of the response.
            case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.STARTS_WITH.value,
            function_arguments={
                "substring": substring,
                "case_sensitive": case_sensitive,
            },
        )

class EndsWith(FunctionEvaluator):
    def __init__(self, substring: str, case_sensitive:bool = False):
        """
        Initialize the EndsWith function evaluator.

        Args:
            substring (str): The substring to check for at the end of the response.
            case_sensitive (bool, optional): If True, the comparison is case-sensitive. Defaults to False.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.ENDS_WITH.value,
            function_arguments={
                "substring": substring,
                "case_sensitive": case_sensitive,
            },
        )

class LengthLessThan(FunctionEvaluator):
    def __init__(self, max_length: int):
        """
        Initialize the LengthLessThan function evaluator.

        Args:
            max_length (int): The maximum length that the response should have.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.LENGTH_LESS_THAN.value,
            function_arguments={"max_length": max_length, },
        )

class LengthGreaterThan(FunctionEvaluator):
    def __init__(self, min_length: int):
        """
        Initialize the LengthGreaterThan function evaluator.

        Args:
            min_length (int): The minimum length that the response should have.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.LENGTH_GREATER_THAN.value,
            function_arguments={"min_length": min_length, },
        )

class ApiCall(FunctionEvaluator):
    def __init__(self, url: str, payload: dict = None, headers: dict = None):
        """
        Initialize the ApiCall function evaluator.

        Args:
            url (str): The URL to make the API call to.
            payload (dict): The payload to be sent in the API call. Response will be added to the dict as "response".
            headers (dict, optional): The headers to be included in the API call. Defaults to None.
        """
        super().__init__(
            function_name=FunctionEvalTypeId.API_CALL.value,
            function_arguments={
                "url": url,
                "payload": payload,
                "headers": headers,
            },
        )


from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import List, Optional
from athina.helpers.logger import logger
from athina.helpers.athina_logging_helper import AthinaLoggingHelper
from athina.interfaces.athina import AthinaExperiment

from athina.interfaces.data import DataPoint
from athina.interfaces.result import BatchRunResult, EvalResult
from athina.services.athina_api_service import AthinaApiService
import traceback


class BaseEvaluator(ABC):
    _experiment: Optional[AthinaExperiment] = None
    # Abstract properties
    @property
    @abstractmethod
    def name(self) -> str:
        """A unique name identifier for the evaluator."""
        pass

    @property
    @abstractmethod
    def display_name(self) -> str:
        """A display name for the evaluator."""
        pass

    @property
    @abstractmethod
    def metric_ids(self) -> List[str]:
        """The metric computed by the evaluator."""
        pass

    @property
    @abstractmethod
    def required_args(self) -> List[str]:
        """A list of required arguments for the evaluator."""
        pass

    @property
    @abstractmethod
    def examples(self):
        """A list of examples for the evaluator."""
        pass

    @abstractmethod
    def is_failure(self, *args) -> Optional[bool]:
        """A method to determine if the evaluation failed."""
        pass

    # Common methods
    def _examples_str(self) -> str:
        return "" if self.examples is None else "\n".join(map(str, self.examples))
    
    def configure_experiment(self, experiment: AthinaExperiment):
        """Configured metadata parameters to log an experiment to Athina"""
        self._experiment = experiment
        return self

    def validate_args(self, **kwargs) -> None:
        """
        Validates that all required arguments are present and not None.
        """
        for arg in self.required_args:
            if arg not in kwargs:
                raise ValueError(f"Missing required argument: {arg}")
            elif kwargs[arg] is None:
                raise ValueError(f"{arg} cannot be None")

    def _validate_batch_args(self, data: List[DataPoint]) -> bool:
        """
        Validates that each entry in the batch has all the required arguments,
        and none of the arguments is None.
        """
        for i, entry in enumerate(data):
            for arg in self.required_args:
                if arg not in entry:
                    raise ValueError(
                        f"Data at index {i} is missing required argument: {arg}"
                    )
                elif entry[arg] is None:
                    raise ValueError(
                        f"Data at index {i} has required argument {arg} set to None"
                    )
        return True

    def _log_evaluation_request(self, data) -> Optional[str]:
        """
        Logs usage to Athina for analytics and creates an evaluation request.
        """
        eval_request_id = None
        try:
            eval_request_id = AthinaLoggingHelper.create_eval_request(
                eval_name=self.name, request_data={"data": data}, request_type="batch"
            )
            self._log_experiment(eval_request_id)
        except Exception as e:
            pass
        return eval_request_id

    def _log_experiment(self, eval_request_id: Optional[str]):
        """
        Logs experiment to Athina if there is an ongoing experiment.
        """
        if self._experiment and eval_request_id:
            AthinaLoggingHelper.log_experiment(
                eval_request_id=eval_request_id,
                experiment=self._experiment,
            )

    def _log_evaluation_results(self, eval_request_id: Optional[str], eval_results: List[EvalResult]):
        """
        Logs the evaluation results to Athina if the eval_request_id is available.
        """
        if eval_request_id:
            try:
                AthinaLoggingHelper.log_eval_results(
                    eval_request_id=eval_request_id,
                    eval_results=eval_results,
                )
            except Exception as e:
                pass

    def run(self, **kwargs) -> BatchRunResult:
        """
        Run the LLM evaluator, and log results to Athina.
        """
        AthinaApiService.log_usage(eval_name=self.name, run_type="batch")
        eval_request_id = self._log_evaluation_request(kwargs)
        eval_result = self._evaluate(**kwargs)
        self._log_evaluation_results(eval_request_id=eval_request_id, eval_results=[eval_result])

        return BatchRunResult(
            eval_request_id=eval_request_id,
            eval_results=[eval_result],
        )
    
    def _run_batch_generator_async(
        self, data: List[DataPoint], max_parallel_evals: int
    ):
        with ThreadPoolExecutor(max_workers=max_parallel_evals) as executor:
            # Submit all tasks to the executor and store them with their original index
            future_to_index = {
                executor.submit(self._evaluate, **entry): i
                for i, entry in enumerate(data)
            }

            # Create a list to store results in the original order
            results = [None] * len(data)

            for future in as_completed(future_to_index):
                index = future_to_index[future]
                try:
                    results[index] = future.result()
                except Exception as e:
                    entry = data[index]
                    logger.error(f"Error running batch async {entry}: {e}")
                    traceback.print_exc()
                    results[index] = None

            return results

    def _run_batch_generator(self, data: List[DataPoint]):
        """
        Generator function for running a batch of evaluations.
        Iterates over a dataset, and runs the evaluator on each entry.
        """
        for entry in data:
            try:
                yield self._evaluate(**entry)
            except Exception as e:
                logger.error(f"Error evaluating entry {entry}: {e}")
                traceback.print_exc()
                yield None

    def run_batch(
        self, data: List[DataPoint], max_parallel_evals: int = 5
    ) -> BatchRunResult:
        """
        Runs the evaluator on a batch of data.
        """
        # Log usage to Athina for analytics
        AthinaApiService.log_usage(eval_name=self.name, run_type="batch")
        eval_request_id = self._log_evaluation_request(data)
        
        # Run the evaluations
        if max_parallel_evals > 1:
            eval_results = self._run_batch_generator_async(data, max_parallel_evals)
        else:
            eval_results = list(self._run_batch_generator(data))

        # Log evaluation results to Athina
        self._log_evaluation_results(eval_request_id, eval_results)

        return BatchRunResult(
            eval_request_id=eval_request_id,
            eval_results=eval_results,
        )


from typing import List, Optional, Dict, Any

from athina.llms.abstract_llm_service import AbstractLlmService
from ..llm_evaluator import LlmEvaluator
from athina.evals.eval_type import LlmEvalTypeId
from ..example import FewShotExample


class CustomPrompt(LlmEvaluator):
    """
    This evaluator can be configured with custom examples and instructions.
    """

    _eval_prompt: Optional[str] = None
    _display_name: str = None
    _metric_ids: List[str] = None
    _model: str = None
    _required_args: List[str] = None
    _examples: List[FewShotExample] = None

    def __init__(
        self,
        eval_prompt: str,
        display_name: str = None,
        metric_ids: List[str] = None,
        model: str = None,
        required_args: List[str] = [],
        examples: List[FewShotExample] = [],
        llm_service: Optional[AbstractLlmService] = None,
        **kwargs
    ):
        if eval_prompt is None:
            raise ValueError("eval_prompt is not defined")
        if model is None:
            raise ValueError("model is not defined")


        self._eval_prompt = eval_prompt
        self._display_name = display_name
        self._metric_ids = metric_ids
        self._model = model
        self._required_args = required_args
        self._examples = examples
        self._system_message_template = self.DEFAULT_SYSTEM_MESSAGE_TEMPLATE + self.RETURN_FORMAT_INSTRUCTIONS

        super().__init__(
            model=self._model,
            system_message_template=self._system_message_template,
            user_message_template=self._eval_prompt,
            llm_service=llm_service,
            **kwargs
        )

    @property
    def name(self):
        return LlmEvalTypeId.CUSTOM_PROMPT.value
        
    @property
    def metric_ids(self) -> List[str]:
        return self._metric_ids
        
    @property
    def display_name(self):
        return self._display_name
        
    @property
    def default_model(self):
        return self._model
        
    @property
    def required_args(self):
        return self._required_args
        
    @property
    def examples(self):
        return self._examples
    
    def is_failure(self, result) -> Optional[bool]:
        return bool(result == "Fail") 
    
    def _user_message(self, **kwargs) -> str:
        return self._user_message_template.format(
            **kwargs,
        )
        



from typing import List, Optional
from athina.interfaces.result import EvalResult, EvalPerformanceReport
from athina.services.athina_api_service import AthinaApiService
from athina.helpers.logger import logger

class EvalPerformanceCalculator:
    """
    Class for calculating the performance metrics of the evaluator.
    """

    @staticmethod
    def calculate_eval_performance_metrics(
        eval_results: List[EvalResult],
        labels: List[bool],
        eval_request_id: Optional[str] = None,
        should_print: bool = False,
        should_log: bool = True,
    ) -> EvalPerformanceReport:
        """
        Calculates and logs the performance metrics for the evaluator.
        """

        if (labels is None) or (len(labels) != len(eval_results)):
            raise ValueError(
                "Labels must be provided and must be the same length as eval_results."
            )

        # Extract predictions from eval_results
        predictions = [result["failure"] for result in eval_results]

        # Initialize counters
        TP, FP, TN, FN = 0, 0, 0, 0

        # Count TP, FP, TN, FN
        for pred, label in zip(predictions, labels):
            if pred == 1 and label == 1:
                TP += 1
            elif pred == 1 and label == 0:
                FP += 1
            elif pred == 0 and label == 0:
                TN += 1
            elif pred == 0 and label == 1:
                FN += 1

        # Calculate metrics
        precision = TP / (TP + FP) if (TP + FP) else 0
        recall = TP / (TP + FN) if (TP + FN) else 0
        accuracy = (TP + TN) / (TP + FP + FN + TN)
        f1_score = (
            2 * (precision * recall) / (precision + recall)
            if (precision + recall)
            else 0
        )
        runtime = int(sum([result["runtime"] for result in eval_results]))
        dataset_size = len(eval_results)

        if should_print:
            logger.info(f"Dataset Size: {dataset_size}")
            logger.info(f"Runtime: {runtime}")
            logger.info(f"True Positives: {TP}")
            logger.info(f"False Positives: {FP}")
            logger.info(f"True Negatives: {TN}")
            logger.info(f"False Negatives: {FN}")
            logger.info(f"Precision: {precision}")
            logger.info(f"Recall: {recall}")
            logger.info(f"Accuracy: {accuracy}")
            logger.info(f"F1 Score: {f1_score}")

        report = EvalPerformanceReport(
            true_positives=TP,
            false_positives=FP,
            true_negatives=TN,
            false_negatives=FN,
            precision=precision,
            recall=recall,
            accuracy=accuracy,
            f1_score=f1_score,
            dataset_size=dataset_size,
            runtime=runtime,
        )

        if eval_request_id is not None and should_log:
            AthinaApiService().log_eval_performance_report(eval_request_id, report)

        return report




from typing import Optional, List
from athina.evals.grounded.similarity import Comparator
from athina.metrics.metric_type import MetricType
import time
from typing import Optional
from athina.interfaces.result import EvalResult, EvalResultMetric
from athina.helpers.logger import logger
from athina.interfaces.athina import AthinaExperiment
from ..base_evaluator import BaseEvaluator

class GroundedEvaluator(BaseEvaluator):

    _comparator: Comparator

    """
    This evaluator runs the requested grounded evaluator on the given data.
    """

    @property
    def _model(self):
        return None
    
    @property
    def name(self):
        return self._comparator.__class__.__name__

    @property
    def display_name(self):
        return self._comparator.__class__.__name__

    @property
    def metric_ids(self) -> List[str]:
        return [MetricType.SIMILARITY_SCORE.value]

    @property
    def examples(self):
        return None

    def __init__(
        self,
        comparator: Comparator = None,
        failure_threshold: Optional[float] = None,
    ):
        if comparator is None:
            raise ValueError(f"comparator is a required argument") 
        else:
            self._comparator = comparator
        if failure_threshold is not None:
            self._failure_threshold = failure_threshold

    def _process_kwargs(self, required_args, **kwargs):
        required_args_map = {key: kwargs[key] for key in required_args}
        if len(required_args_map) == 2:
            values = list(required_args_map.values())
            if all(isinstance(value, str) for value in values):
                string1, string2 = values
                return string1, string2
            else:
                raise ValueError("Both arguments must be strings.")
        else:
            raise ValueError("Exactly two arguments are required.")

    def is_failure(self, score) -> Optional[bool]:
        return bool(score < self._failure_threshold) if self._failure_threshold is not None else None

    def _evaluate(self, **kwargs) -> EvalResult:
        """
        Run the Function evaluator.
        """
        start_time = time.perf_counter()

        # Validate that correct args were passed
        self.validate_args(**kwargs)
        metrics = []
        try: 
            string1, string2 = self._process_kwargs(self.required_args, **kwargs)
            # Calculate the similarity score using the comparator
            similarity_score = self._comparator.compare(string1, string2)
            metrics.append(EvalResultMetric(id=MetricType.SIMILARITY_SCORE.value, value=similarity_score))
            if self._failure_threshold is None:
                explanation = f"Successfully calculated similarity score of {similarity_score} using {self.display_name}"
            elif similarity_bool(score < self._failure_threshold):
                explanation = f"Evaluation failed as similarity score of {similarity_score} is below the failure threshold of {self._failure_threshold} using {self.display_name}"
            else:
                explanation = f"Evaluation succeeded as similarity score of {similarity_score} is above the failure threshold of {self._failure_threshold} using {self.display_name}"

            failure = self.is_failure(similarity_score)
        except Exception as e:
            logger.error(f"Error occurred during eval: {e}")
            raise e

        end_time = time.perf_counter()
        eval_runtime_ms = int((end_time - start_time) * 1000)
        eval_result = EvalResult(
            name=self.name,
            display_name=self.display_name,
            data=kwargs,
            reason=explanation,
            runtime=eval_runtime_ms,
            model=None,
            metrics=metrics,
            failure=failure
        )
        return {k: v for k, v in eval_result.items() if v is not None}

